// AMD Interwave local ROM\DRAM dumper - by wbcbz7 o2.11.2o17

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <conio.h>
#include <ctype.h>
#include <dos.h>
#include "iwdump.h"

FILE          *f;
int           iw_baseport;
unsigned char buffer[BUF_SIZE];
bool          gf1mode = false;

void iwread(unsigned long pos, void *data, unsigned long length, bool rom) {
    unsigned char *buf = (unsigned char*)data;
    unsigned char romflag = (rom ? 2 : 0);
    
    // set dram\rom and enable autoincrement
    outp (iw_baseport+IW_PIDX,   IW_LMCI);
    outp (iw_baseport+IW_PDATAH, ((inp(iw_baseport+IW_PDATAH) & 0xFB) | romflag | 1));
    
    outp (iw_baseport+IW_PIDX,   IW_LMALI);
    outpw(iw_baseport+IW_PDATAL, (pos & 0xFFFF));
    outp (iw_baseport+IW_PIDX,   IW_LMALH);
    outp (iw_baseport+IW_PDATAH, (pos >> 16));
    
    for (int i = 0; i < length; i++) {
        if (gf1mode) {
            // legacy GF1 (GUS Classic/MAX) mode
            outp (iw_baseport+IW_PIDX,   IW_LMALI);
            outpw(iw_baseport+IW_PDATAL, ((pos+i) & 0xFFFF));
            outp (iw_baseport+IW_PIDX,   IW_LMALH);
            outp (iw_baseport+IW_PDATAH, ((pos+i) >> 16));
        }
        buf[i] = inp(iw_baseport+IW_LMBDR);
    }
    
    // set dram as source
    outp (iw_baseport+IW_PIDX,   IW_LMCI);
    outp (iw_baseport+IW_PDATAH, (inp(iw_baseport+IW_PDATAH) & 0xFD));
}

// minimal interwave reset (preserves DRAM)
// return 1 if ok
int iwinit() {
    int i;
    char a;

    // reset card
    outp (iw_baseport+IW_PIDX,   0x4C);
    outp (iw_baseport+IW_PDATAH, 0);
    for (i = 0; i < 40; i++) {
        inp(iw_baseport+IW_LMBDR);
    }
    outp (iw_baseport+IW_PIDX,   0x4C);
    outp (iw_baseport+IW_PDATAH, 1);
    for (i = 0; i < 40; i++) {
        inp(iw_baseport+IW_LMBDR);
    }
    
    // try to set enhanced mode
    outp (iw_baseport+IW_PIDX,   0x99);
    a = inp(iw_baseport+IW_PDATAH);
    outp (iw_baseport+IW_PIDX,   0x19);
    outp (iw_baseport+IW_PDATAH, (a | 1));
    
    outp (iw_baseport+IW_PIDX,   0x99);
    a = inp(iw_baseport+IW_PDATAH) & 1;
    return a;
}

int iwgetport() {
    char *env = getenv("ULTRASND");
    int  port, dummy;
    
    if (!env) return 0;
    sscanf(env, "%x,%d,%d,%d,%d", &port, &dummy, &dummy, &dummy, &dummy);
    return port;
}

typedef struct {
    long   start;
    long   len;
    char   fname[32];   // enough for dos :)
} memsize_t;

memsize_t rom = {0, 0, ROM_FILE}, ram = {0, 0, RAM_FILE};

void parse(int argc, char *argv[]) {
    bool curmode = true;               // rom first
    char tmp[1024] = {0};
    
    for (int i = 1; i < argc; i++) {
        strncpy(tmp, argv[i], sizeof(char)*1024); strupr(tmp);
        
        // legacy gf1 mode
        if ((strstr(tmp, "GF1")) || (strstr(tmp, "LEGACY")))  {gf1mode = true; continue;}
        
        // mode switch
        if (!strcmp(tmp, "ROM")) {rom.len = TOTAL_MEM; curmode = true; continue;}
        if (!strcmp(tmp, "RAM")) {ram.len = TOTAL_MEM; curmode = false; continue;}
        
        char *tok = strtok(tmp, "-=");
        if (!strcmp(tok, "FILE")) {
            tok = strtok(NULL, "-=");
            
            strncpy((curmode ? rom.fname : ram.fname), tok, sizeof(char)*32);
            continue;
        }
        if (!strcmp(tok, "START")) {
            tok = strtok(NULL, "-=");
            long sz; sscanf(tok, "%x", &sz);
            
            if (curmode) rom.start = sz; else ram.start = sz;
            continue;
        }
        if ((!strcmp(tok, "SIZE")) || (strstr(tok, "LEN"))) {
            tok = strtok(NULL, "-=");
            long sz;      sscanf(tok, "%d", &sz);
            long mul = 1;
            
            if (strstr(tok, "K")) mul = 1024;
            if (strstr(tok, "M")) mul = 1024*1024;
            if (curmode) rom.len = sz*mul; else ram.len = sz*mul;
            continue;
        }
    }
}

int main(int argc, char *argv[]) {
    long i;
    
    printf("AMD Interwave local ROM\\DRAM dumper - by wbcbz7 o1.11.2o17\n");
    
    parse(argc, argv);
    
    if ((argc < 2) || ((!rom.len) && (!ram.len)))  {
        printf("usage:   iwdump [rom/ram] [file=<fname>] [start=<hex offset>] [size=<bytes>]\n\n");
        printf("example: iwdump rom start=100000 size=1M\n");
        printf("         - dump 1MB  of rom from 2nd MB to IWROM.BIN\n");
        printf("example: iwdump ram file=dump.bin size=32K\n");
        printf("         - dump 32KB of ram from start  to DUMP.BIN\n");
        printf("example: iwdump rom file=rom.bin size=512 start=10000 ram \n");
        printf("         - dump 512 bytes of rom from 64k offset to DUMP.BIN and entire\n");
        printf("           ROM space to IWDRAM.bin\n");
        printf("example: iwdump gf1 ram file=dump.bin size=1M\n");
        printf("         - dump 1MB of ram from start   to DUMP.BIN in GUS Classic/MAX mode\n");
        return 0;
    }
    
    
    // init
    iw_baseport = iwgetport();
    if (iw_baseport == 0) {printf("error: can't get Interwave base port - check ULTRASND environment varable\n"); return 1;}
    
    if (!iwinit()) {
        printf("error: can't init Interwave chip, continue in GF1 (GUS Classic/MAX) mode? <Y/N>\n");
        char ch; do {ch = toupper(getch());} while (!strchr("YN\x1A", ch)); if (ch != 'Y') return 1;
        gf1mode = true;
    }
    
    // dump
    if (rom.len) {
        long pos = rom.start;
        printf("dumping ROM content to %s, please wait...\n", rom.fname);
        f = fopen(rom.fname, "wb");
        for (i = 0; i < (rom.len / BUF_SIZE); i++) {
            iwread(pos, &buffer, BUF_SIZE, true);
            fwrite(&buffer, sizeof(unsigned char), BUF_SIZE, f);
            pos += BUF_SIZE;
            if (!(i & 0x3)) {printf("."); fflush(stdout);}
        }
        // remainder
        if ((pos - rom.start) < rom.len) {
            iwread(pos, &buffer, rom.len + rom.start - pos, false);
            fwrite(&buffer, sizeof(unsigned char), rom.len + rom.start - pos, f);
        }
        fclose(f);
        printf("\n-> %d bytes from 0x%X ok\n", rom.len, rom.start);
    }   
    
    if (ram.len) {
        long pos = ram.start;
        printf("dumping DRAM content to %s, please wait...\n", ram.fname);
        f = fopen(ram.fname, "wb");
        for (i = 0; i < (ram.len / BUF_SIZE); i++) {
            iwread(pos, &buffer, BUF_SIZE, false);
            fwrite(&buffer, sizeof(unsigned char), BUF_SIZE, f);
            pos += BUF_SIZE;
            if (!(i & 0x3)) {printf("."); fflush(stdout);}
        }
        // remainder
        if ((pos - ram.start) < ram.len) {
            iwread(pos, &buffer, ram.len + ram.start - pos, false);
            fwrite(&buffer, sizeof(unsigned char), ram.len + ram.start - pos, f);
        }
        fclose(f);
        printf("\n-> %d bytes from 0x%X ok\n", ram.len, ram.start);
    }
    
    printf("done\n");
}
